#include <stdio.h>
#include <stdlib.h> 

/* This unit implements a data structure for one-dimensional dynamic semi-infinite
   range minimum queries for keys in range [0...n-1]. Query time is O(log n). 
   The structure is a complete binary tree with n leaves supporting 
   predecessor and successor queries, and insertions and deletions.

   Range query [-infty,x) is answered by calling i=predecessor(x) and
   taking value l(i) of i from an external array.

   Inserting a point x is done by calling insert(x) and then calling
   deleteGreaterSuccessors(x,values), which removes all successors
   of x that have greater value than x. This guarantees that range
   query [-infty,x) is answered correctly.*/

typedef struct {
   unsigned int mantissa2 : 32;
   unsigned int mantissa1 : 20;
   unsigned int exponentbias1023 : 11;
   unsigned int signbit : 1;
} doubleMask;

typedef union {
   double asDouble;
   doubleMask asMask;
} doubleAndMask;

inline unsigned int log_2(unsigned long n) {
   doubleAndMask fm;
   fm.asDouble = (double) n;
   return (fm.asMask.exponentbias1023 - 1023);
}

class treeNode {
public:
   bool isLeft;
   bool isRight;
}; 

class completeBinaryTree {
private:
   unsigned int leaves;
   treeNode *A;
public: 
   completeBinaryTree(unsigned int leaves) 
   {
      unsigned int i;
      this->leaves = leaves;  /* !!! leaves must be 2^x for some x */
      A = new treeNode[2*leaves]; /* only indices 1...2*leaves are used */
      /* no leaves in the tree at first */
      for (i=0;i<2*leaves;i++) {
         A[i].isLeft = 0;
         A[i].isRight = 0;
      }
   };
   ~completeBinaryTree() 
   {
      delete A;
   };
   unsigned int predecessor(unsigned int index)
   {
      /* search the predecessor of i in the tree */
   
      /* in normal tree presentation:
         // x = "ith leaf"
         while (x!=NULL && (x==x->parent->left || !x->parent->isLeft))
            x = x->parent;
         // x->parent is the LCA of the original x and predesessor(x)	 
         if (x==NULL) return leafs;
         x = x->parent->left;
         while (!isLeaf(x)) 
            if (x->isRight) x = x->right;
    	    else x = x->left;
         return x; */	
   
      /* in array implementation: */
      unsigned int i = leaves+index;
      while (i>1 && ((i==(i>>1)<<1) || !A[i/2].isLeft))
         i = i/2;
      if (i==1) return leaves;
      i = i-1;
      while (i<leaves)
         if (A[i].isRight) i = 2*i+1;
         else i = 2*i;
      return i-leaves;             
   }; 
   unsigned int successor(unsigned int index)
   {
      /* search the successor of i in the tree */
   
      /* in normal tree presentation:
         // x = "ith leaf"
         while (x!=NULL && (x==x->parent->right || !x->parent->isRight))
            x = x->parent;
         // x->parent is the LCA of the original x and successor(x)	 
         if (x==NULL) return leafs;
         x = x->parent->right;
         while (!isLeaf(x)) 
            if (x->isLeft) x = x->left;
   	    else x = x->right;
         return x; */	
   
      /* in array implementation: */
      unsigned int i = leaves+index;
      while (i>1 && ((i!=(i>>1)<<1) || !A[i/2].isRight))
         i = i/2;
      if (i==1) return leaves;
      i = i+1;
      while (i<leaves)
         if (A[i].isLeft) i = 2*i;
         else i = 2*i+1;
      return i-leaves;             
   }; 
   void insertLeaf(unsigned int index)
   {
      /* insert leaf 'index' and update information to the root */
      unsigned int i = leaves+index;
      while (i>1) {
         if (i==(i>>1)<<1) if (A[i/2].isLeft) break; else A[i/2].isLeft = 1;
         else if (A[i/2].isRight) break; else A[i/2].isRight = 1;
         i = i/2;
      }
   };
   void deleteLeaf(unsigned int index)
   {
      /* delete leaf 'index' and update information to the root */
      unsigned int i = leaves+index;
      while (i>1 && !A[i].isLeft && !A[i].isRight) {
         if (i==(i>>1)<<1) A[i/2].isLeft = 0;
         else A[i/2].isRight = 0;
         i = i/2;
      }
   }; 
   void deleteGreaterSuccessors(unsigned int index, int *values)
   {
      /* delete j=successor(index) until values[j]<values[index] */ 
      unsigned int j = successor(index);
      while (j<leaves && values[j]>=values[index]) {
         deleteLeaf(j);
         j = successor(index);
      }
   };    
}; 


/* testing...
int main(int argc, char** argv)  
{
    unsigned int n = atoi(argv[1]);
    unsigned int log2n = log_2(n);    
    unsigned int i;
    unsigned int leaves = 1<<(log2n+1);
    int *values= new int[leaves];
    printf("log_2(%d)=%d, %d<2^%d=%d\n",n,log2n,n,log2n+1,leaves);
    srand(n);
    completeBinaryTree *T = new completeBinaryTree(leaves);    
    for (i=leaves;i>0;i--) {
       T->insertLeaf(i-1);  
       values[i-1] = rand() % 100;
       T->deleteGreaterSuccessors(i-1,values);
    } 
    // print leaves of tree 
    for (i=0;i<leaves;i++)
       printf("values[%d]=%d,",i,values[i]);
    printf("\n");
    i=0;
    while (i<leaves) {
       printf("values[%d]=%d>",i,values[i]);
       i = T->successor(i);
    }      
    printf("-1\n");
    delete T;
    delete values;
    return 1;
}*/
