
/* This unit implements a dynamic data structure for one-dimensional 
   range minimum queries using a red-black tree as a basis. Each subtree 
   stores the minima of its left and right subtrees. Queries are answered 
   almost like in range trees, except that points are stored in internal 
   nodes, not in leafs.  

   The extra complication is that the minima need to be maintained
   during insertions, deletions, and rebalancing. This is accomplished
   by some extra updates during these operations; all operations still
   take O(log n) time. */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <time.h>


#define compEQ(a,b) (a.i == b.i && a.j == b.j)
#define compLT(a,b) (a.i < b.i || (a.i==b.i && a.j<b.j))
#define compLEQ(a,b) (a.i < b.i || (a.i==b.i && a.j<=b.j))
#define compGT(a,b) compLEQ(b,a)
#define compGEQ(a,b) compLT(b,a)

#define min2(a,b) (a<b?a:b)
#define min3(a,b,c) (a<b?(a<c?a:c):(b<c?b:c))

/* implementation dependent declarations */
typedef enum {
   STATUS_OK,
   STATUS_MEM_EXHAUSTED,
   STATUS_DUPLICATE_KEY,
   STATUS_KEY_NOT_FOUND
} statusEnum;

class keyType {
public:
   int i; // primary key
   int j; // secondary key
};

/* user data stored in tree */
class recType {
public:
   int v; // label of a point
   int minl; // minimum label in the left subtree
   int minr; // minimum label in the right subtree
};

/* implementation independent declarations */
/* Red-Black tree description */
typedef enum { BLACK, RED } nodeColor;

class nodeType {
public:
    nodeType *left;       /* left child */
    nodeType *right;      /* right child */
    nodeType *parent;     /* parent */
    nodeColor color;            /* node color (BLACK, RED) */
    keyType key;                /* key used for searching */
    recType rec;                /* user data */
};


#define NIL &sentinel           /* all leafs are sentinels */

static nodeType sentinel = { NIL, NIL, 0, BLACK, 0, 0, INT_MAX, INT_MAX, INT_MAX};

class redBlackTree {
private:
   /* last node found, optimizes find/delete operations */
   nodeType *lastFind;
   /* root of Red-Black tree */
   nodeType *root;    
   void updatePathMin(nodeType *x);   
   void updateParentMin(nodeType *x);   
   void rotateLeft(nodeType *x);   
   void rotateRight(nodeType *x);
   void insertFixup(nodeType *x);
   void deleteFixup(nodeType *x);   
   statusEnum find(keyType key, recType *rec);     
   int updateMinimaRec(nodeType *current);
   void makeEmptyRec(nodeType *current);      
public:
   redBlackTree() {root = NIL;};
   ~redBlackTree() {makeEmpty();};
   statusEnum insertNode(keyType key, recType *rec);
   statusEnum deleteNode(keyType key);
   void updateMinimaAll(); // for testing the correctness of dynamic updates
   int rangeMin(keyType l, keyType r);   
   void makeEmpty();   
};

void redBlackTree::updatePathMin(nodeType *x) {

   /*************************************************   
    * updates subtree minima in path from x to root *
    *************************************************/ 

    while (x != root) {
      if (x == x->parent->left) 
         x->parent->rec.minl = min3(x->rec.v,x->rec.minl,x->rec.minr);
      else
         x->parent->rec.minr = min3(x->rec.v,x->rec.minl,x->rec.minr);
      x = x->parent;
    }
}

void redBlackTree::updateParentMin(nodeType *x) {

   /******************************************   
    * updates subtree minima for parent of x *
    ******************************************/ 

    if (x != root)
      if (x == x->parent->left) 
         x->parent->rec.minl = min3(x->rec.v,x->rec.minl,x->rec.minr);
      else
         x->parent->rec.minr = min3(x->rec.v,x->rec.minl,x->rec.minr);
}


void redBlackTree::rotateLeft(nodeType *x) {

   /**************************
    *  rotate node x to left *
    **************************/

    nodeType *y = x->right;

    /* establish x->right link */
    x->right = y->left;
    /* update minimum in the right subtree of x */
    x->rec.minr = y->rec.minl;
    /* update minimum in the left subtree (to come) of y */
    y->rec.minl = min3(x->rec.v,x->rec.minl,x->rec.minr);

    if (y->left != NIL) y->left->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->left) {
            x->parent->left = y;
            /* update minimum in the left subtree of x->parent */
            x->parent->rec.minl = min3(y->rec.v,y->rec.minl,y->rec.minr);
        }
        else {
            x->parent->right = y;
            /* update minimum in the right subtree of x->parent */
            x->parent->rec.minr = min3(y->rec.v,y->rec.minl,y->rec.minr);
        }
    } else {
        root = y;
    }

    /* link x and y */
    y->left = x;
    if (x != NIL) x->parent = y;
}

void redBlackTree::rotateRight(nodeType *x) {

   /****************************
    *  rotate node x to right  *
    ****************************/

    nodeType *y = x->left;

    /* establish x->left link */
    x->left = y->right;
    /* update minimum in the left subtree of x */
    x->rec.minl = y->rec.minr;
    /* update minimum in the right subtree (to come) of y */
    y->rec.minr = min3(x->rec.v,x->rec.minl,x->rec.minr);

    if (y->right != NIL) y->right->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->right) {
            x->parent->right = y;
            /* update minimum in the right subtree of x->parent */
            x->parent->rec.minr = min3(y->rec.v,y->rec.minl,y->rec.minr);
        }
        else {
            x->parent->left = y;
            /* update minimum in the left subtree of x->parent */
            x->parent->rec.minl = min3(y->rec.v,y->rec.minl,y->rec.minr);
        }
    } else {
        root = y;
    }

    /* link x and y */
    y->right = x;
    if (x != NIL) x->parent = y;
}

void redBlackTree::insertFixup(nodeType *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after inserting node x           *
    *************************************/

    /* No changes needed here for maintaining the subtree minima,
       except in rotation procedures */

    /* check Red-Black properties */
    while (x != root && x->parent->color == RED) {
        /* we have a violation */
        if (x->parent == x->parent->parent->left) {
            nodeType *y = x->parent->parent->right;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->right) {
                    /* make x a left child */
                    x = x->parent;
                    rotateLeft(x);
                }

                /* recolor and rotate */
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateRight(x->parent->parent);
            }
        } else {

            /* mirror image of above code */
            nodeType *y = x->parent->parent->left;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->left) {
                    x = x->parent;
                    rotateRight(x);
                }
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateLeft(x->parent->parent);
            }
        }
    }
    root->color = BLACK;
}

statusEnum redBlackTree::insertNode(keyType key, recType *rec) {
    nodeType *current, *parent, *x;

   /***********************************************
    *  allocate node for data and insert in tree  *
    ***********************************************/

    /* find future parent */
    current = root;
    parent = 0;
    while (current != NIL) {
        if (compEQ(key, current->key)) 
            return STATUS_DUPLICATE_KEY;
        parent = current;
        current = compLT(key, current->key) ?
            current->left : current->right;
    }

    /* setup new node */
    if ((x = new nodeType) == 0)
        return STATUS_MEM_EXHAUSTED;
    x->parent = parent;
    x->left = NIL;
    x->right = NIL; 
    x->color = RED;
    x->key = key;
    x->rec = *rec;

    /* insert node in tree */
    if(parent) {
        if(compLT(key, parent->key)) {
            parent->left = x;
            /* update minimum in the left subtree */
            parent->rec.minl = x->rec.v; 
        }
        else {
            parent->right = x;
            /* update minimum in the right subtree */
            parent->rec.minr = x->rec.v; 
        }
    } else {
        root = x;
    }

    insertFixup(x);

    /* It is now enough to update all values from x to the root.
       The nodes that got wrong values during rebalancing,
       will be corrected by this step, since they are
       are on the way from x to the root */
    updatePathMin(x);

    lastFind = NULL;

    return STATUS_OK;
}

void redBlackTree::deleteFixup(nodeType *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after deleting node x            *
    *************************************/
    /* No changes needed here for maintaining the subtree minima,
       except in rotation procedures */
    while (x != root && x->color == BLACK) {
        if (x == x->parent->left) {
            nodeType *w = x->parent->right;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateLeft (x->parent);
                w = x->parent->right;
            }
            if (w->left->color == BLACK && w->right->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->right->color == BLACK) {
                    w->left->color = BLACK;
                    w->color = RED;
                    rotateRight (w);
                    w = x->parent->right;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->right->color = BLACK;
                rotateLeft (x->parent);
                x = root;
            }
        } else {
            nodeType *w = x->parent->left;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateRight (x->parent);
                w = x->parent->left;
            }
            if (w->right->color == BLACK && w->left->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->left->color == BLACK) {
                    w->right->color = BLACK;
                    w->color = RED;
                    rotateLeft (w);
                    w = x->parent->left;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->left->color = BLACK;
                rotateRight (x->parent);
                x = root;
            }
        }
    }
    x->color = BLACK;
}

statusEnum redBlackTree::deleteNode(keyType key) {
    nodeType *x, *y, *z;

   /*****************************
    *  delete node z from tree  *
    *****************************/

    /* find node in tree */
    if (lastFind && compEQ(lastFind->key, key))
        /* if we just found node, use pointer */
        z = lastFind;
    else {
        z = root;
        while(z != NIL) {
            if(compEQ(key, z->key)) 
                break;
            else
                z = compLT(key, z->key) ? z->left : z->right;
        }
        if (z == NIL) return STATUS_KEY_NOT_FOUND;
    }

    if (z->left == NIL || z->right == NIL) {
        /* y has a NIL node as a child */
        y = z;
    } else {
        /* find tree successor with a NIL node as a child */
        y = z->right;
        while (y->left != NIL) y = y->left;
    }

    /* x is y's only child */
    if (y->left != NIL)
        x = y->left;
    else
        x = y->right;

    /* remove y from the parent chain */
    x->parent = y->parent;
    if (y->parent)
        if (y == y->parent->left) 
            y->parent->left = x;
        else
            y->parent->right = x;
    else
        root = x;

    if (y != z) {
        z->key = y->key;
        z->rec.v = y->rec.v;
        /* Left minimum of z->parent may change.
           This effect does not need to be carried out to the root,
           since it will be taken into account later by updatePath(x) */
        updateParentMin(z);
    }

    if (y->color == BLACK)
        deleteFixup(x);

    /* It is now enough to update all values from x to the root.
       The nodes that got wrong values during rebalancing,
       will be corrected by this step, since they are
       are on the way from x to the root */
    updatePathMin(x);

    delete y;
    lastFind = NULL;

    return STATUS_OK;
}

statusEnum redBlackTree::find(keyType key, recType *rec) {

   /*******************************
    *  find node containing data  *
    *******************************/

    nodeType *current = root;
    while(current != NIL) {
        if(compEQ(key, current->key)) {
            *rec = current->rec;
            lastFind = current;
            return STATUS_OK;
        } else {
            current = compLT (key, current->key) ?
                current->left : current->right;
        }
    }
    return STATUS_KEY_NOT_FOUND;
}


int redBlackTree::updateMinimaRec(nodeType *current) {

    /*******************************
     *  update minima in subtrees  *
     *******************************/
    if (current->left != NIL)
       current->rec.minl = updateMinimaRec(current->left);
    if (current->right != NIL)
       current->rec.minr = updateMinimaRec(current->right);
    
    /* this would print the tree in the postorder
    if (!current->parent)
       printf("root(%d,%d,%d,%d)",current->key.i,current->rec.v,current->rec.minl,current->rec.minr);
    else if (current == current->parent->left)
       printf("lc(%d,%d,%d,%d)",current->key.i,current->rec.v,current->rec.minl,current->rec.minr);
    else if (current == current->parent->right)
       printf("rc(%d,%d,%d,%d)",current->key.i,current->rec.v,current->rec.minl,current->rec.minr);
    fflush(stdout);
    */
    
    return min3(current->rec.v,current->rec.minl,current->rec.minr); 
}

void redBlackTree::updateMinimaAll() {
   updateMinimaRec(root);
}

int redBlackTree::rangeMin(keyType l, keyType r) {

   /*******************************
    *  find minimum from a range  *
    *******************************/

    int min = INT_MAX;
    nodeType *limit;

    /* find the split-node first */
    nodeType *current = root;
    while(current != NIL && (compLEQ(r,current->key) || compGT(l,current->key))) {
        if (compLEQ(l,current->key) && compGEQ(r,current->key)) min = min2(min,current->rec.v);
        if (compLEQ(r, current->key))
            current = current->left;
        else 
            current = current->right;
    }
    if (current == NIL)
       return min; /* no split-node found ==> no points inside the query range */
      
    /* report minimum in path to $l$ */
    limit = current->left;
    min = current->rec.v;
    while(limit != NIL) {
       if (compLEQ(l, limit->key)) {
           min = min3(min,limit->rec.minr,limit->rec.v);
           limit = limit->left;
       }
       else  
          limit = limit->right;
    }
    /* report minimum in path to $r$ */
    limit = current->right;
    while(limit != NIL) {
       if (compGEQ(r, limit->key)) {
           min = min3(min,limit->rec.minl,limit->rec.v);
           limit = limit->right;
       }
       else  
          limit = limit->left;
    }
    return min;
}

void redBlackTree::makeEmpty()
{
   makeEmptyRec(root);
   root = NIL;
}

void redBlackTree::makeEmptyRec(nodeType *current)
{
   if (current->left != NIL) makeEmptyRec(current->left);
   if (current->right != NIL) makeEmptyRec(current->right);
   delete current;    
}


/* testing the structure
void main(int argc, char **argv) {
    int i;
    recType rec;
    keyType key;
    statusEnum status;
    keyType left, right;
    int keys[1000]; 
    int values[10000]; 
    time_t t;
    redBlackTree *T = new redBlackTree();

    left.i = atoi(argv[1]);
    right.i = atoi(argv[2]);
    left.j = INT_MIN;
    right.j = INT_MAX;

    srand((unsigned)time(&t));
    for(i=0; i<1000; i++) {
       rec.v = rand()%10000; key.i = rand() % 1000;  key.j = i; rec.minl = INT_MAX; rec.minr = INT_MAX;  
       status = T->insertNode(key,&rec);
       keys[i] = key.i;
       values[rec.v] = i;
    }       
    printf("Minimum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMin(left,right));
    T->updateMinimaAll();
    printf("Minimum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMin(left,right));
    for(i=0; i<100; i++) {
       if (T->rangeMin(left,right)==INT_MAX) break;
       key.j = values[T->rangeMin(left,right)];
       key.i = keys[key.j];   
       status = T->deleteNode(key);
    }   
    if (T->rangeMin(left,right)<INT_MAX) 
       printf("Minimum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMin(left,right));
    else
       printf("No points in range [%d,%d]\n",left.i,right.i);
    fflush(stdout);
    T->updateMinimaAll();
    if (T->rangeMin(left,right)<INT_MAX) 
       printf("Minimum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMin(left,right));
    else
       printf("No points in range [%d,%d]\n",left.i,right.i);
    delete T;  
}*/
