
/* This code implements sparse dynamic programming algoritms for edit distance, 
   using range minimum queries that are implemented using a red-black tree.   
   The approach is compared with the naive DP solutions. 
   See Makinen, Navarro, Ukkonen: "Transposition invariant string matching", 
   STACS 2003, for details.  */ 



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include "386.c"
#include "rbt.cpp"
#include "1drangequery.cpp"

#define max2(a,b) ((a)>(b)?(a):(b))
//#define min2(a,b) ((a)<(b)?(a):(b))

class integerNode {
public:
   integerNode *next;
   int x;  
};

class integerList {
public:
   integerNode *first;
   integerNode *last;
};

class occNode {
public:
   occNode *next;
   int x;  
   int t;
};

class occList {
public:
   occNode *first;
   occNode *last;
};

class keyTypeNode {
public:
   keyType key;
   keyTypeNode *next;
};

class matchList {
public:
   keyTypeNode *first;
   keyTypeNode *last;
};

/* The set of matching character pairs M is implemented as a list */
static matchList M;

/* Assuming integer alphabet Sigma,
   each symbol keeps a list of its occurrences in A */
static integerList Sigma[256];

/* Match set for each transposition -256,...,256 (mapped to 0...512) */
static matchList Mt[513];


int preProcess(unsigned char *A,unsigned char *B, int m, int n, matchList *M,
               integerList Sigma[], int t) {
   integerNode *node,*temp;

   /*******************************************************************
    * this produces the match set M in reverse column-by-column order *
    *******************************************************************/

   int i,j,k;
   for (k=0;k<256;k++) {
      Sigma[k].first = NULL;
      Sigma[k].last = NULL;
   }
   for (i=m;i>=1;i--) { 
      node = new integerNode; 
      node->next = NULL; 
      node->x = i;
      if (Sigma[(unsigned int)A[i]].last != NULL) {
         Sigma[(unsigned int)A[i]].last->next = node;
         Sigma[(unsigned int)A[i]].last = node;
      }
      else {
         Sigma[(unsigned int)A[i]].first = node;
         Sigma[(unsigned int)A[i]].last = node; 
      }      
   } 
   /* let's add (0,0) to the match set */
   M->first = new keyTypeNode;
   M->last = M->first;
   M->last->next = NULL;
   M->last->key.i = 0;
   M->last->key.j = 0;
   k = 1;
   for (j=1;j<=n;j++) {
      if  ((int)B[j]-t<0 || (int)B[j]-t >255) continue;
      node = Sigma[(unsigned int)B[j]-t].first;    
      while (node != NULL) {
         M->last->next = new keyTypeNode;
         M->last = M->last->next;
         M->last->next = NULL;
         M->last->key.i = node->x;
         M->last->key.j = j;
	 node = node->next;
         k++; 
      } 
   } 
   /* let's add (m+1,n+1) */
   M->last->next = new keyTypeNode;
   M->last = M->last->next;
   M->last->next = NULL;
   M->last->key.i = m+1;
   M->last->key.j = n+1;
   /* let's return the size of M */
   return k;
}

void preProcessAllTranspositions(unsigned char *A,unsigned char *B, int m, int n, matchList Mt[]) {

   /****************************************************************************
    * this produces all the match sets M_{t} in reverse column-by-column order *
    ****************************************************************************/

   int i,j,t;
   for (t=0; t<513; t++) {
      Mt[t].first = new keyTypeNode;
      Mt[t].last = Mt[t].first;
      Mt[t].last->next = NULL;
      Mt[t].last->key.i = 0;
      Mt[t].last->key.j = 0;
   }
   for (j=1; j<=n; j++) 
      for (i=m; i>=1; i--) {
         t = (int)B[j]-(int)A[i]+256;
	 Mt[t].last->next = new keyTypeNode;
         Mt[t].last = Mt[t].last->next;
         Mt[t].last->next = NULL;
         Mt[t].last->key.i = i;
         Mt[t].last->key.j = j;	 
      }   
   /* let's add (m+1,n+1) */
   for (t=0; t<513; t++) {
      Mt[t].last->next = new keyTypeNode;
      Mt[t].last = Mt[t].last->next;
      Mt[t].last->next = NULL;
      Mt[t].last->key.i = m+1;
      Mt[t].last->key.j = n+1;	 
   }
   return;
}


int processSparse(int m, int n, matchList *M) {
   recType rec;
   keyType left, right;
   keyTypeNode *match; 
   redBlackTree *T = new redBlackTree();
   
   /**************************************************************** 
    * computes d_{ID}(A+t,B) using one-dimensional range searching *
    ****************************************************************/

   // time complexity O(|M|\log (|M|))
   rec.v = 0; rec.minl = INT_MAX; rec.minr = INT_MAX;
   left.i = 0; left.j = 0; right.j = INT_MAX;
   T->insertNode(M->first->key, &rec); 
   match = M->first->next;
   while (match != NULL) {
      right.i = match->key.i-1;
      rec.v = T->rangeMin(left,right)-2;        
      T->insertNode(match->key,&rec);  
      match = match->next;
   }
   delete T;
   return rec.v+m+n+2; 
}

int processSparseFast(int m, int n, matchList *M) {
   keyTypeNode *match; 
   int i,d;
   int *values = new int[m+2];
   unsigned int leaves = 1<<(log_2(m+1)+1);
   completeBinaryTree *T = new completeBinaryTree(leaves);
   
   /**************************************************************** 
    * computes d_{ID}(A+t,B) using one-dimensional range searching *
    ****************************************************************/
    
   // time complexity O(|M|\log m)
   values[0] = 0;
   T->insertLeaf(0); 
   for (i=1;i<=m+1;i++)
      values[i]=INT_MAX;
   match = M->first->next;
   while (match != NULL) {
      i = T->predecessor(match->key.i);
      T->insertLeaf(match->key.i); 
      if (values[i]-2<values[match->key.i]) {
         values[match->key.i] = values[i]-2;
	 T->deleteGreaterSuccessors(match->key.i,values);
      }	 
      match = match->next;
   }
   d = values[m+1]+m+n+2;
   delete values;
   delete T;
   return d;
}

void searchOccurrences(int m, int n, int k, int t, matchList *M, int *socc, int
*eocc) {
   keyTypeNode *match; 
   int i,j,d,count, occ_count;
   int *values = new int[m+2]; 
   unsigned int leaves = 1<<(log_2(m)+1);
   completeBinaryTree *T = new completeBinaryTree(leaves);
   
   /**************************************************************** 
    * reports {j} such that d_{ID}(A+t,T_{j'...j})<=k              *
    ****************************************************************/
    
   // time complexity O(|M|\log m)
   values[0] = 0;
   T->insertLeaf(0); 
   for (i=1;i<=m+1;i++)
      values[i]=INT_MAX;
   match = M->first->next;
   // discard the pair (m+1,n+1) since it is only used in the distance
   // computation (hence using match->next != NULL)
   while (match->next != NULL) {
      i = T->predecessor(match->key.i);
      // let's check if cheeper to start a new occurrence
      d = min2(values[i]-2,-match->key.j-1); //-key.j-1+key.i+key.j=key.i-1
      T->insertLeaf(match->key.i); 
      if (d<values[match->key.i]) {
         values[match->key.i] = d;
	 T->deleteGreaterSuccessors(match->key.i,values);
      }	 
      // report an interval [key.j,j] on the last row, 
      // where the current point induces occurrences d_{ID}(A+t,T_{key.j...j})<=k
      j = k-d-m; //d+key.i+key.j+(m-key.i)+(j-key.j)=k
      //if (match->key.i == m && match->key.j == m) printf("d=%d j=%d k=%d t=%d\n",d,j,k,t);
      if (j >= match->key.j) {
         socc[match->key.j]++;
	 eocc[min2(j,n)]++;
      }
      match = match->next;
   }
   delete values;
   delete T;
}

int reportOccurrences(int n, int *socc, int *eocc, occList *L) {
   int count = 0, occ_count = 0,j;
   // collect the list of occurrences
   for (j=1; j<=n; j++) {
      count += socc[j];
      if (count > 0) {
         occ_count++;
         L->last->next = new occNode;
	 L->last = L->last->next;
	 L->last->x = j;
	 L->last->t = 0;
	 L->last->next = NULL;
      }
      count -= eocc[j];
   }
   return occ_count;
}


int processSparseAlpha(unsigned char *B,int m, int n, matchList *M, integerList Sigma[], int alpha) {
   recType rec;
   keyType left, right, key;
   keyTypeNode *match; 
   integerNode *temp;
   int oldj;
   redBlackTree *T = new redBlackTree();
   
   /************************************************************************* 
    * computes d_{ID}^{\alpha}(A+t,B) using one-dimensional range searching *
    *************************************************************************/

   rec.v = 0; rec.minl = INT_MAX; rec.minr = INT_MAX;
   left.i = 0; left.j = 0; right.j = INT_MAX;
   T->insertNode(M->first->key, &rec); 
   match = M->first->next;
   oldj=0;
   while (match != NULL) {
      if (match->key.j>oldj && match->key.j-alpha-2>=0) { 
         /* need to remove column j-alpha-2 */
         if (match->key.j-alpha-2==0) {
            /* cell (0,0) separately */
            key.i = 0; key.j = 0;
            T->deleteNode(key);
         }
         else {       
            temp = Sigma[(unsigned int)B[match->key.j-alpha-2]].first;
            while (temp!=NULL) {
               key.i = temp->x; key.j = match->key.j-alpha-2;
               T->deleteNode(key);
               temp = temp->next;
            } 
         }
      }
      right.i = match->key.i-1; left.i = match->key.i-alpha-1;
      rec.v = T->rangeMin(left,right)-2;        
      T->insertNode(match->key,&rec);  
      oldj = match->key.j;
      match = match->next;
   }
   return rec.v+m+n+2; 
}

int processDP(unsigned char *A,unsigned char *B, int m, int n, int **d, int t) {
   int i,j,v;

   /**********************************************************
    * computes d_{ID}(A+t,B) using basic dynamic programming *
    **********************************************************/

   for (i=0;i<m+1;i++)
      d[i][0] = i;
   for (j=0;j<n+1;j++)
      d[0][j] = j;
   for (i=1;i<m+1;i++)
      for (j=1;j<n+1;j++)
         d[i][j] = min3(d[i-1][j-1]+((int)A[i]+t==(int)B[j]?0:2),d[i-1][j]+1,d[i][j-1]+1);
   return d[m][n];
}

void searchDP(unsigned char *A,unsigned char *T, int m, int n, int k, int **d,
int t, int *socc, int *eocc) {
   int i,j,v,occ=0;

   /**********************************************************************
    * searches d_{ID}(A+t,T_{j'...j})<=k using basic dynamic programming *
    **********************************************************************/

   for (i=0;i<m+1;i++)
      d[i][0] = i;
   for (j=0;j<n+1;j++)
      d[0][j] = 0;
   for (i=1;i<m+1;i++)
      for (j=1;j<n+1;j++)
         d[i][j] = min3(d[i-1][j-1]+((int)A[i]+t==(int)T[j]?0:2),d[i-1][j]+1,d[i][j-1]+1);
   for (j=1; j<=n; j++)
      if (d[m][j]<=k) {socc[j]++;eocc[j]++;}
}

int processDPAlpha(int m, int n, matchList *M, int **d, int alpha) {
   int i,j,min;
   keyTypeNode *match;

   /*******************************************************************
    * computes d_{ID}^{\alpha}(A+t,B) using basic dynamic programming *
    *******************************************************************/
   
   for (i=0;i<m+1;i++)
      for (j=0;j<n+1;j++)
         d[i][j] = INT_MAX;

   d[0][0] = 0;
   match = M->first->next;   
   while (match != NULL) {
      min = INT_MAX;
      for (i=max2(0,match->key.i-alpha-1); i<match->key.i;i++)
         for (j=max2(0,match->key.j-alpha-1); j<match->key.j;j++)
            min = min2(min,d[i][j]);
      d[match->key.i][match->key.j] = min-2;
      match = match->next;
   }
   return d[m+1][n+1]+m+n+2;
}

void emptyOcclist(occList *L) {
   occNode *temp=L->first;
   while (temp!=NULL) {
      L->first = temp->next;
      delete temp;
      temp = L->first;
   }
   delete L;
}

void cleanUp(int **d, int dim, matchList *M, matchList Mt[]) {
   int i,j;
   integerNode *node, *temp;
   keyTypeNode *knode,*ktemp;
   for (i=0;i<256;i++) {
      node = Sigma[i].first;
      while (node!=NULL) {
         temp = node;
         node = node->next;
         delete temp; 
      }
   }
   for (i=0;i<dim;i++)
      delete d[i];
   delete d;
   knode = M->first;
   while (knode != NULL) {
      ktemp = knode;
      knode = knode->next;
      delete ktemp;
   }
   for (i=0;i<513;i++) {
      knode = Mt[i].first;
      while (knode!=NULL) {
         ktemp = knode;
         knode = knode->next;
         delete ktemp;
      }
   }
   return;
} 

unsigned char* readFile(char *name, int *size) {
   FILE *input;
   long len;
   unsigned char *S;  

   input = fopen(name,"r");
   fseek(input, 0L, SEEK_END);  
   len = ftell(input);     
   rewind(input);               
   S = (unsigned char*)calloc(len + 2, sizeof(unsigned char));
   /* read file in S starting from S[1] */
   fread(S+1, len, 1, input); 
   fclose(input);
   *size = len;
   return S;
}

int main(int argc, char **argv) {
    int i,t=0,alpha,K,sparse=0,sparseFast,dp,sparseAlpha,dpAlpha;
    long tPreProcess,tPreProcessAll,tSparse=0,tSparseFast,tDp,tSparseAlpha, tDpAlpha, tProcess;
    int min;
    recType rec;
    keyType key;
    statusEnum status;
    keyType left, right;
    
    /* strings A and B */
    unsigned char *A,*B;
    /* size of A and B */
    int m,n,k;
    /* Dynamic programming matrix */
    int **d;
    
    if (argc!=5) {
       printf("Syntax: %s k alpha fileA fileB\n",argv[0]);
       return 1;
    }

    k = atoi(argv[1]);
    alpha = atoi(argv[2]);
    A = readFile(argv[3],&m);
    B = readFile(argv[4],&n);
    
    printf("|A|=%d,|B|=%d\n",m,n);
    fflush(stdout);  

    startclock();
    K = preProcess(A,B,m,n,&M,Sigma,t);
    tPreProcess = stopclock();

    printf("|M|=%d\n",K);
    fflush(stdout);  

    startclock();
    sparse = processSparse(m,n,&M);
    tSparse = stopclock();
    
    startclock();
    sparseFast = processSparseFast(m,n,&M);
    tSparseFast = stopclock();
    
    startclock();
    sparseAlpha = processSparseAlpha(B,m,n,&M,Sigma,alpha);
    tSparseAlpha = stopclock();
    
    d = (int**)malloc(sizeof(int*)*(m+2));
    for (i=0;i<m+2;i++)
       d[i] = (int*)malloc(sizeof(int)*(n+2));

    startclock();
    dp = processDP(A,B,m,n,d,t); 
    tDp = stopclock();

    startclock();
    dpAlpha = processDPAlpha(m,n,&M,d,alpha); 
    tDpAlpha = stopclock();


    printf("\nd_{ID}(A,B) = %d (sparse), =%d (sparseFast), = %d (DP)\n\nd_{ID}^{\\alpha=%d} = %d (sparse), = %d (DP)\n",sparse,sparseFast,dp,alpha,sparseAlpha,dpAlpha);
    printf("\npreprocessing took %dms, sparse %dms, sparseFast %dms, DP %dms,sparseAlpha %dms, DPAlpha %dms\n",tPreProcess,tSparse,tSparseFast,tDp,tSparseAlpha,tDpAlpha);       
    fflush(stdout);

    /* Finally let's try out transpositions */
    
    startclock();
    preProcessAllTranspositions(A,B,m,n,Mt);
    tPreProcessAll = stopclock();
    
    printf("\nPreprocessing for all transpositions took %dms\n",tPreProcessAll);
           
    startclock();
    min = INT_MAX;
    for (t=0;t<513;t++) {
       if (Mt[t].first->next == Mt[t].last) continue;
       sparseFast = processSparseFast(m,n,&(Mt[t]));
       min = min2(min,sparseFast);
    }
    tProcess = stopclock();    
    
    printf("\nSparse(fast) DP took %dms, d_{ID}^{t}(A,B)=%d\n",tProcess,min);
    
    startclock();
    min = INT_MAX;
    occList *L = new occList;
    L->first = new occNode;
    L->last = L->first;
    L->first->next = NULL;
    int *socc = new int[n+1], *eocc = new int[n+1];
    int j;
    for (j=1; j<=n; j++) {
       socc[j] = 0;
       eocc[j] = 0;
    }   
    for (t=0;t<513;t++) {
       if (Mt[t].first->next == Mt[t].last) continue;
       searchOccurrences(m,n,k,t-256,&(Mt[t]),socc,eocc);
    }
    int occ = reportOccurrences(n,socc,eocc,L);
    tProcess = stopclock();    
    
    printf("\nSearch(fast) took %dms, found %d occurrences\n",tProcess,occ);
    emptyOcclist(L);
    
    startclock();
    min = INT_MAX;
    for (t=0;t<513;t++) {
       if (Mt[t].first->next == Mt[t].last) continue;
       sparse = processSparse(m,n,&(Mt[t]));
       min = min2(min,sparse);
    }
    tProcess = stopclock();    
    
    printf("\nSparse DP took %dms, d_{ID}^{t}(A,B)=%d\n",tProcess,min);

    startclock();
    min = INT_MAX;
    for (t=0;t<513;t++) {
       if (Mt[t].first->next == Mt[t].last) continue;
       dp = processDP(A,B,m,n,d,t-256);   
       min = min2(min,dp);
    }
    tProcess = stopclock();    
    
    printf("\nDense DP took %dms, d_{ID}^{t}(A,B)=%d\n",tProcess,min);
    
    startclock();
    min = INT_MAX;
    occ = 0;
    L = new occList;
    L->first = new occNode;
    L->first->next = NULL;
    for (j=1; j<=n; j++) {
       socc[j] = 0;
       eocc[j] = 0;
    }   
    for (t=0;t<513;t++) {
       if (Mt[t].first->next == Mt[t].last) continue;
       searchDP(A,B,m,n,k,d,t-256,socc,eocc);   
    }
    occ = reportOccurrences(n,socc,eocc,L);
    tProcess = stopclock();    
    
    printf("\nSearch took %dms, found %d occurrences\n",tProcess,occ);
    emptyOcclist(L);
    
    cleanUp(d,m+2,&M,Mt);
    delete socc;
    delete eocc;
    return 1;
}
